#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <glib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <rtp/rtp-packet.h>
#include <rtp/rtcp-packet.h>
/*#include <rtp/rtp-audio.h>*/

#include "rtp.h"

#define BUFF_SIZE 10

static const gchar App_Name[] = { /* Name for APP packets */
  'p', 'v'
};

int
find_host(gchar *host_name, struct in_addr *address)
{
  struct hostent *host;

  if (inet_aton(host_name, address)) {
    return 1;
  } else { 	
    /* if user has given a hostname not an IP */
    host = gethostbyname(host_name);
  }
  
  if (!host) {
    g_warning( "error looking up host: %s", host_name );
    return 0;
  }

  memcpy(address, (struct in_addr *) host->h_addr_list[0],
         sizeof(struct in_addr));

  return 1;
}

void
rtp_send( GSList *sen_list,
	 gpointer buf, int nbytes, 
	 rtp_payload_t pt, guint32 nsamp)
{
  Rtp_Packet packet;
  GSList *iterator;
  rtp_sender_info *sen;
  

  packet = rtp_packet_new_allocate(nbytes, 0, 0);

  rtp_packet_set_csrc_count(packet, 0);
  rtp_packet_set_extension(packet, 0);
  rtp_packet_set_padding(packet, 0);
  rtp_packet_set_version(packet, RTP_VERSION);
  rtp_packet_set_payload_type(packet, ( guint8 )pt ) ;
  rtp_packet_set_marker(packet, 0);


  g_memmove(rtp_packet_get_payload(packet), buf, nbytes);
  
  for( iterator = sen_list; iterator; iterator = g_slist_next( iterator ) ) {
	sen = (rtp_sender_info *)iterator->data;
      
  	rtp_packet_set_ssrc(packet, g_htonl( sen->ssrc ) ); 
  	rtp_packet_set_seq(packet, g_htons( sen->seq ) );
  	rtp_packet_set_timestamp(packet, g_htonl( sen->timestamp ) );
	
	//g_print( "sending packet with seq. #%d", sen->seq );
	sen->addr.sin_port = g_htons(sen->port);
	rtp_packet_send(packet, sen->rtp_sock, (struct sockaddr *)&sen->addr, 
		  		sizeof( struct sockaddr_in ) );
  	++sen->seq;
  	sen->timestamp += nsamp;
	++sen->packets_sent;
	sen->octets_sent += rtp_packet_get_packet_len(packet);
  }
  //g_print( "sent rtp packet\n" );
  rtp_packet_free(packet);
}

Rtp_Packet
rtp_really_receive( rtp_receiver_info* recv_info, guint32 mtu )
{
  Rtp_Packet packet;
  guint32 packet_size ;
  struct sockaddr_in from_addr ;
  socklen_t from_len = sizeof( from_addr );
  
  packet = rtp_packet_read( recv_info->rtp_sock,
			          (struct sockaddr *)&from_addr,
		       	          &from_len );
  packet_size = rtp_packet_get_payload_len( packet ) ;
  
  if( packet_size > mtu )
  g_warning( "Recieved RTP Packet's size is larger than MTU\n" ) ; 
  
  ++recv_info->packets_sent;
  recv_info->octets_sent += packet_size;

  return packet;
}

Rtp_Packet
rtp_receive( rtp_receiver_info* recv_info, guint32 mtu )
{
  /* I am including a packet sequencing code of mine, if you dont like
  *  it, plz uncomment the commented code & delete the rest including
  *  all unused varriables. 
  */
  static Rtp_Packet packets[ BUFF_SIZE ] ;
  static gboolean buffer_full = FALSE ;
  static gboolean first_time  = TRUE ;
  static gboolean missed = FALSE ;
  static gint missing = 0 ;
  static gint last = 0 ;
  
  Rtp_Packet packet;
  guint16 seq;
  guint8 i;
  //guint32 packet_size ;
  //struct sockaddr_in from_addr ;
  //socklen_t from_len = sizeof( from_addr );
  
  if( !missed ) {
      packet = rtp_really_receive( recv_info, mtu ) ; 
      seq = g_ntohs( rtp_packet_get_seq( packet ) ) ;
      
      //g_print( "no miss %d\n", seq ) ;

      if( first_time ) {
      	recv_info->seq = seq ;
      	first_time = FALSE ;
     	return packet ;
      } 
  
      else if( recv_info->seq+1 != seq ) {
        g_print( "seq %d %d\n", recv_info->seq, seq ) ;
	missed  = TRUE ;
	missing = (recv_info->seq+1 ) % BUFF_SIZE ;
	last    = (recv_info->seq   ) % BUFF_SIZE ;
	
	for( i = 0 ; i < BUFF_SIZE ; i++ )
	packets[ i ] = NULL ;

	packets[ seq % BUFF_SIZE ] = packet ;
	return NULL ;
      }

      else {
	recv_info->seq = seq ;
	return packet ;
      }
  }

  else {
    //g_print( "miss %d %d\n", missing, last ) ;
    
    if( buffer_full ) {
       //g_print( "buffer full\n" ) ;
       if( missing == last ) {
	 buffer_full = FALSE ;
	 missed = FALSE ;
	 recv_info->seq = g_ntohs( rtp_packet_get_seq( packets[ last ] ) );
	 return packets[ last ] ;
       }
       else {
	 if( missing >= BUFF_SIZE )
	 missing = 0 ;
	 return packets[ missing++ ] ;
       }
    }
    else {
        packet = rtp_really_receive( recv_info, mtu ) ;
        seq = g_ntohs( rtp_packet_get_seq( packet ) );
 
        if( seq % BUFF_SIZE == last )
	   buffer_full = TRUE ;
	
	//g_print( "%d\n", seq ) ;
	packets[ seq % BUFF_SIZE ] = packet ;
	return NULL ;
    }
  }

 /* packet = rtp_packet_read( recv_info->rtp_sock,
			          (struct sockaddr *)&from_addr,
		       	          &from_len );
  packet_size = rtp_packet_get_payload_len( packet ) ;
  
  if( packet_size > mtu )
  g_warning( "Recieved RTP Packet's size is larger than MTU\n" ) ; 
  
  g_print( "Recieved RTP Packet no.%d\n", 
		  g_ntohs( rtp_packet_get_seq( packet ) ) ); 
  
  ++recv_info->packets_sent;
  recv_info->octets_sent += packet_size ;

  return packet ;*/
}

void
rtcp_send( GSList *sen_list, Rtcp_APP_message message ) 
{
  Rtcp_Compound compound;
  GSList *iterator;
  rtp_sender_info *sen;

  compound = rtcp_compound_new_allocate(RTP_MTU);
  rtcp_compound_add_sr(compound, sen->ssrc, sen->timestamp,
                       sen->packets_sent, sen->octets_sent );
  message.packets_sent = g_htonl(sen->packets_sent );
  
  rtcp_compound_add_app(compound, sen->ssrc, App_Name,
                        &message, sizeof message ) ;
  
  for( iterator = sen_list; iterator ; iterator = g_slist_next( iterator ) ) {
	sen = (rtp_sender_info *)iterator->data;

      	sen->addr.sin_port = g_htons(sen->port+1);
      	g_print("Sending rtcp message to %s...\n", sen->hostname );
  
      	rtcp_compound_send(compound, sen->rtp_sock, 
		    (struct sockaddr *)&sen->addr, 
		     sizeof( struct sockaddr_in ) );
      	sen->addr.sin_port = g_htons(sen->port);
  }
  
  g_print( "sent rtcp compound\n" );
  rtcp_compound_free(compound);
}

void
parse_rtcp_packet( Rtcp_Packet packet, Rtcp_APP_message* message )
{
  switch( rtcp_packet_get_packet_type( packet ) ) 
  {
    case RTCP_BYE:
    /* We should have been recieving this type of Rtcp Packet but the 
    *  rtcp_send sends an APP Packet in an RTCP Compound & i couldnt 
    *  find any method in rtcp-packet.h( librtp header file ) or gphone
    *  source code that could tell me how to add a BYE packet to the RTCP 
    *  Compound. If you know that plz inform me
    */ 
	break;
    
    case RTCP_APP:
     	g_print( "RTCP APP recieved\n" );
	g_memmove( message, 
		   rtcp_app_packet_get_data( packet ), sizeof( *message ) ) ;
        break;
   
    case RTCP_SR:
     	g_print( "RTCP SR recieved\n" );
        break;
    
    default:
     	g_print( "Unknown RTCP Packet recieved\n" );
        break;
  }
}

Rtcp_APP_message*
rtcp_receive( rtp_receiver_info* recv )
{
  Rtcp_Compound compound;
  struct sockaddr_in from_addr;
  socklen_t from_len ;
  Rtcp_APP_message* message = g_malloc( sizeof( Rtcp_APP_message ) ) ;

  from_len = sizeof from_addr;
  compound = rtcp_compound_read( recv->rtcp_sock,
		                 (struct sockaddr *)&from_addr, 
		  		 &from_len ) ;
  rtcp_compound_foreach( compound, ( Rtcp_Foreach_Func )parse_rtcp_packet, 
		  	 				         message ) ;
  rtcp_compound_free(compound);

  return message ;
}

void
rtp_client_connection( GSList **sen_list) {
  if( *sen_list != NULL ) {
  }

  else {
       rtp_sender_info *sen = g_malloc( sizeof( rtp_sender_info ) );
  
       g_print( "initializing socket\n" ) ;
       if ((sen->rtp_sock = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
          g_print("open_sockets : Send_Sock");
       }
       
       *sen_list = g_slist_append( *sen_list, sen); 
  }
}

void
rtp_server_connection( rtp_receiver_info *recv ) {
  /* RTP Socket */
  if ((recv->rtp_sock = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
    g_print("open_sockets : RTP Sock");
  }
  
  if ( bind( recv->rtp_sock, ( struct sockaddr *)&recv->addr,
	     sizeof( recv->addr ) ) < 0) {
    g_print("bind_sockets : RTP Sock\n");
  }

  /* RTCP Socket */
  recv->addr.sin_port = g_htons(recv->port+1);
  if ((recv->rtcp_sock = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
    g_print("open_sockets : RTCP Sock");
  }
  
  if ( bind( recv->rtcp_sock, ( struct sockaddr *)&recv->addr,
	     sizeof( recv->addr ) ) < 0) {
    g_print("bind_sockets : RTCP Sock\n");
    g_print("%d\n", sizeof( recv->addr ) );
  }
  recv->addr.sin_port = g_htons(recv->port);
}

int
rtp_connection_call( GSList **sen_list, 
		     gchar *hostname,
		     guint port)
{
  struct in_addr call_host;     /* host info of computer to call */
  rtp_sender_info *sen; 
  guint8 ret_val = 0;

  if( *sen_list == NULL ) {
	return ret_val;
  }
  else if( g_slist_length( *sen_list ) == 1 ) {
	sen = (rtp_sender_info *)(*sen_list)->data ;
  }
  else {
	rtp_sender_info *tmp = (rtp_sender_info *)g_slist_nth( *sen_list, 0 );
	sen = g_malloc( sizeof( rtp_sender_info ) );
      	sen->rtp_sock = tmp->rtp_sock;
  }
  
  if( ( ret_val = find_host(hostname, &call_host ) ) ) {
   	g_print( "found host %s\n", hostname );

	sen->addr.sin_family = AF_INET;
  	sen->addr.sin_port = g_htons( port );
    	g_memmove(&sen->addr.sin_addr.s_addr,
              	&call_host, sizeof (struct in_addr));
    
	sen->seq = random32(1);
    	sen->ssrc = random32(2);
    	sen->timestamp = random32(3);
    	sen->hostname = g_strdup(hostname);
    	sen->port = port;
    	sen->packets_sent = 0;
  
	if( g_slist_length( *sen_list ) != 1 ) {
    	    *sen_list = g_slist_append( *sen_list, sen );
	}
  }
  return ret_val;
}

int
rtp_server_init( rtp_receiver_info *recv, guint port )
{
  recv->addr.sin_addr.s_addr = htonl( INADDR_ANY ) ;
  recv->addr.sin_family = AF_INET;
  recv->addr.sin_port = g_htons( port );
  
  recv->seq = 0 ; 
  recv->ssrc = random32(2);
  recv->timestamp = random32(3);
  recv->hostname = g_strdup( "localhost" ) ;
  recv->port = port ;
  recv->packets_sent = 0;
  
  return 1;
}

void
rtp_close_all( GSList **sen_list )
{
    GSList *iterator;
    rtp_sender_info *sen;

    if( *sen_list != NULL ) {
    	for( iterator = *sen_list; 
			iterator; iterator = g_slist_next( iterator ) ) {
	     sen = (rtp_sender_info *)iterator->data;

	     close( sen->rtp_sock ) ;
	     g_free( sen->hostname );
	     g_free( sen );
	}
    	g_slist_free( *sen_list );
    	*sen_list = NULL;
    }
}
